from pydfs_lineup_optimizer.sites.fantasy_draft.importer import *
from pydfs_lineup_optimizer.sites.fantasy_draft.settings import *


__all__ = [
    'FantasyDraftCSVImporter', 'FantasyDraftBasketballSettings', 'FantasyDraftFootballSettings',
    'FantasyDraftHockeySettings', 'FantasyDraftBaseballSettings', 'FantasyDraftGolfSettings',
]



import csv
from copy import deepcopy
from itertools import islice
from typing import List, Dict
from pydfs_lineup_optimizer.exceptions import LineupOptimizerIncorrectCSV
from pydfs_lineup_optimizer.lineup_importer import CSVImporter
from pydfs_lineup_optimizer.player import Player, GameInfo
from pydfs_lineup_optimizer.lineup import Lineup, LineupPlayer


class FanDuelCSVImporter(CSVImporter):  # pragma: nocover
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._games = {}

    def _row_to_player(self, row: Dict) -> Player:
        try:
            away_team, home_team = row.get('Game', '').split('@')
            game_info = self._games.get((home_team, away_team))
            if not game_info:
                game_info = GameInfo(home_team, away_team, None)
                self._games[(home_team, away_team)] = game_info
        except ValueError:
            game_info = None
        try:
            player = Player(
                row['Id'],
                row['First Name'],
                row['Last Name'],
                row['Position'].split('/'),
                row['Team'],
                float(row['Salary']),
                float(row['FPPG'] or 0),
                is_injured=True if row['Injury Indicator'].strip() else False,
                game_info=game_info,
                **self.get_player_extra(row)
            )
        except KeyError:
            raise LineupOptimizerIncorrectCSV
        return player

    def import_players(self) -> List[Player]:
        self._games = {}
        with open(self.filename, 'r') as csv_file:
            start_line = 0  # Find line with 'FPPG', that's where players data starts
            while True:
                line = csv_file.readline()
                if 'FPPG' in line:
                    csv_file.seek(0)
                    csv_data = csv.DictReader(islice(csv_file, start_line, None),
                                              skipinitialspace=True)
                    return [self._row_to_player(row) for row in csv_data]
                elif line == '':
                    raise LineupOptimizerIncorrectCSV
                else:
                    start_line += 1

    def import_lineups(self, players):
        with open(self.filename, 'r') as csv_file:
            lines = csv.reader(csv_file)
            try:
                header = next(lines)
                start_column = 3  # First 3 columns has info about tournament
                end_column = header.index('Instructions') - 1
            except (IndexError, ValueError):
                raise LineupOptimizerIncorrectCSV
            position_names = header[start_column:end_column]
            players_dict = {player.id: player for player in players}
            lineups = []
            for line in lines:
                if not line[0]:
                    break
                lineup_players = []
                for index, position in zip(range(start_column, end_column), position_names):
                    player_id = line[index]
                    try:
                        player = players_dict[player_id]
                    except KeyError:
                        raise LineupOptimizerIncorrectCSV('Player not found in players pool')
                    lineup_players.append(LineupPlayer(player, position))
                lineups.append(Lineup(lineup_players))
            return lineups


class FanDuelMVPCSVImporter(FanDuelCSVImporter):
    def import_players(self):
        players = super().import_players()
        mvps = []
        for player in players:
            mvp_player = deepcopy(player)
            mvp_player.fppg *= 1.5
            mvp_player._original_positions = player.positions
            mvp_player.positions = ('MVP', )
            mvps.append(mvp_player)
        players.extend(mvps)
        return players


class FanDuelLOLCSVImporter(FanDuelCSVImporter):
    def import_players(self):
        players = super().import_players()
        stars = []
        for player in players:
            if 'TEAM' in player.positions:
                continue
            star_player = deepcopy(player)
            star_player.fppg *= 1.5
            star_player._original_positions = player.positions
            star_player.positions = ('STAR', )
            stars.append(star_player)
        players.extend(stars)
        return players



from typing import List
from pydfs_lineup_optimizer.settings import BaseSettings, LineupPosition
from pydfs_lineup_optimizer.constants import Sport, Site
from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry
from pydfs_lineup_optimizer.lineup_printer import DraftKingTiersLineupPrinter
from pydfs_lineup_optimizer.sites.draftkings.tiers.importer import DraftKingsTiersCSVImporter
from pydfs_lineup_optimizer.rules import DraftKingsTiersRule


class DraftKingsTiersSettings(BaseSettings):
    site = Site.DRAFTKINGS_TIERS
    csv_importer = DraftKingsTiersCSVImporter
    min_games = 2
    budget = None
    positions = []  # type: List[LineupPosition]
    extra_rules = [DraftKingsTiersRule]
    lineup_printer = DraftKingTiersLineupPrinter


@SitesRegistry.register_settings
class DraftKingsTiersBasketballSettings(DraftKingsTiersSettings):
    sport = Sport.BASKETBALL


@SitesRegistry.register_settings
class DraftKingsTiersBaseballSettings(DraftKingsTiersSettings):
    sport = Sport.BASEBALL


@SitesRegistry.register_settings
class DraftKingsTiersHockeySettings(DraftKingsTiersSettings):
    sport = Sport.HOCKEY


@SitesRegistry.register_settings
class DraftKingsTiersFootballSettings(DraftKingsTiersSettings):
    sport = Sport.FOOTBALL



import os
from typing import Type
from pydfs_lineup_optimizer.solvers.base import Solver
from pydfs_lineup_optimizer.solvers.pulp_solver import PuLPSolver
from pydfs_lineup_optimizer.solvers.constants import SolverSign
from pydfs_lineup_optimizer.solvers.exceptions import SolverException, SolverInfeasibleSolutionException


__all__ = ['Solver', 'PuLPSolver', 'SolverSign', 'SolverException', 'SolverInfeasibleSolutionException',
           'get_default_solver']


def get_default_solver() -> Type[Solver]:
    solver_backend = os.environ.get('SOLVER_BACKEND')
    if solver_backend is None:
        return PuLPSolver
    solver_backend_name = solver_backend.lower()
    if solver_backend_name == 'pulp':
        return PuLPSolver
    elif solver_backend_name == 'mip':
        from pydfs_lineup_optimizer.solvers.mip_solver import MIPSolver
        return MIPSolver
    raise ValueError('Unknown solver backend: %s' % solver_backend)



from pydfs_lineup_optimizer.exceptions import LineupOptimizerIncorrectCSV
from pydfs_lineup_optimizer.player import Player
from pydfs_lineup_optimizer.sites.draftkings.classic.importer import DraftKingsCSVImporter


class DraftKingsCaptainModeCSVImporter(DraftKingsCSVImporter):  # pragma: nocover
    def _row_to_player(self, row):
        try:
            fppg_multiplier = 1.5 if row['Roster Position'] == 'CPT' else 1
            name = row['Name'].split(maxsplit=1)
            player = Player(
                row['ID'],
                name[0],
                name[1] if len(name) > 1 else '',
                row['Roster Position'].split('/'),
                row['TeamAbbrev'],
                float(row['Salary']),
                float(row['AvgPointsPerGame']) * fppg_multiplier,
                game_info=self._parse_game_info(row),
                **self.get_player_extra(row)
            )
        except KeyError:
            raise LineupOptimizerIncorrectCSV
        return player



from typing import FrozenSet, List, Optional, Set, Union, Iterable, Dict, Literal, DefaultDict
from collections import defaultdict
from itertools import chain
from pydfs_lineup_optimizer.settings import BaseSettings
from pydfs_lineup_optimizer.player import Player, GameInfo
from pydfs_lineup_optimizer.exceptions import LineupOptimizerException
from pydfs_lineup_optimizer.utils import link_players_with_positions, ratio, list_intersection
from pydfs_lineup_optimizer.settings import LineupPosition


DirtyPlayer = Union[Player, str]


class BaseFilter:
    def filter(self, player: Player, include_not_matched: bool = True) -> bool:
        raise NotImplemented


class PlayerFilter(BaseFilter):
    def __init__(
            self,
            from_value: float = 0,
            to_value: Optional[float] = None,
            positions: Optional[Iterable[str]] = None,
            teams: Optional[Iterable[str]] = None,
            filter_by: Literal['fppg', 'efficiency', 'salary'] = 'fppg',
    ):
        self.from_value = from_value
        self.to_value = to_value
        self.positions = frozenset(positions or [])
        self.teams = frozenset(teams or [])
        self.filter_by = filter_by

    def filter(self, player, include_not_matched: bool = True) -> bool:
        if self.teams and player.team not in self.teams:
            return include_not_matched
        for position in self.positions:
            if position and position not in player.positions:
                return include_not_matched
        player_attr = getattr(player, self.filter_by)
        if player_attr < self.from_value:
            return False
        if self.to_value and player_attr > self.to_value:
            return False
        return True


class PlayerPool:
    def __init__(self, settings: BaseSettings):
        self._settings = settings
        self._remaining_positions = settings.positions[:]
        self._players: List[Player] = []
        self._players_by_name: DefaultDict[str, List[Player]] = defaultdict(list)
        self._players_by_id: Dict[str, Player] = {}
        self._exclude_teams: Set[str] = set()
        self._locked_players: Dict[Player, Optional[LineupPosition]] = {}
        self._player_filters: List[BaseFilter] = []
        self.with_injured = False
        self.search_threshold = 0.8
        self.removed_players: Set[Player] = set()

    @property
    def all_players(self) -> List[Player]:
        return self._players

    @property
    def filtered_players(self) -> List[Player]:
        pool = []
        for player in self._players:
            if (
                (player in self.removed_players) or
                (player.max_exposure is not None and player.max_exposure <= 0) or
                (not self.with_injured and player.is_injured) or
                player.team in self._exclude_teams or
                (self._player_filters and
                 not all(filter.filter(player) for filter in self._player_filters))
            ):
                continue
            pool.append(player)
        return pool

    @property
    def locked_players(self) -> List[Player]:
        return list(self._locked_players.keys())

    @property
    def locked_positions(self) -> List[LineupPosition]:
        return list(pos for pos in self._locked_players.values() if pos)

    @property
    def locked_players_with_positions(self) -> Dict[Player, LineupPosition]:
        return {player: position for player, position in self._locked_players.items() if position}

    @property
    def available_teams(self) -> FrozenSet[str]:
        return frozenset(player.team for player in self._players)

    @property
    def available_positions(self) -> FrozenSet[str]:
        return frozenset(chain.from_iterable(
            position.positions for position in self._settings.positions))

    @property
    def available_roster_spots(self) -> FrozenSet[str]:
        return frozenset(pos.name for pos in self._settings.positions)

    @property
    def total_players(self) -> int:
        return self._settings.get_total_players()

    @property
    def remaining_players(self) -> int:
        return self.total_players - len(self.locked_players)

    @property
    def remaining_positions(self) -> List[LineupPosition]:
        return self._remaining_positions

    @property
    def games(self) -> FrozenSet[GameInfo]:
        return frozenset(player.game_info for player in self.filtered_players if player.game_info)

    @property
    def budget(self) -> Optional[float]:
        return self._settings.budget

    @property
    def used_budget(self) -> float:
        return sum(player.salary for player in self.locked_players)

    @property
    def remaining_budget(self) -> Optional[float]:
        if self.budget is None:
            return None
        return self.budget - self.used_budget

    def reset_players(self) -> None:
        self._players = []
        self._players_by_name = defaultdict(list)
        self._players_by_id = {}
        self._exclude_teams = set()
        self.removed_players = set()
        self._locked_players = {}
        self._player_filters = []

    def reset_locked(self) -> None:
        self._locked_players = {}

    def load_players(self, players: Iterable[Player]) -> None:
        self.reset_players()
        self.extend_players(players)

    def extend_players(self, players: Iterable[Player]):
        for player in players:
            self.add_player(player)

    def add_player(self, player: Player):
        self._players.append(player)
        self._players_by_name[player.full_name].append(player)
        self._players_by_id[player.id] = player

    def get_player_by_name(
            self, player_name: str, position: Optional[str] = None,
            allowed_players: Optional[Set[Player]] = None,
    ) -> Optional[Player]:
        players = set(self._players_by_name.get(player_name, set()))
        if position:
            players = {player for player in players if position in player.positions}
        if allowed_players:
            players = players.intersection(allowed_players)
        if players:
            if len(players) > 1:
                raise LineupOptimizerException('More than 1 player is found for: %s' % player_name)
            return list(players)[0]
        if self.search_threshold:
            matched_players = self.get_players(PlayerFilter(positions=[position])) if position else self.all_players
            possibilities = [(player, ratio(player_name, player.full_name)) for player in matched_players]
            filtered_possibilities = filter(lambda pos: pos[1] >= self.search_threshold, possibilities)
            if filtered_possibilities:
                sorted_players = sorted(filtered_possibilities, key=lambda pos: -pos[1])
                return next(map(lambda player: player[0], sorted_players))  # typing: ignore
        return None

    def get_player_by_id(self, player_id: str) -> Optional[Player]:
        return self._players_by_id.get(player_id)

    def exclude_teams(self, teams: Iterable[str]):
        self._exclude_teams = set(teams)

    def remove_player(self, player: DirtyPlayer):
        self.removed_players.add(self._clean_player(player))

    def restore_player(self, player: DirtyPlayer):
        try:
            self.removed_players.remove(self._clean_player(player))
        except KeyError:
            raise LineupOptimizerException('Player not removed!')

    def lock_player(self, player: DirtyPlayer, position: Optional[str] = None):
        player = self._clean_player(player)
        if player.max_exposure == 0:
            raise LineupOptimizerException('Can\'t add this player to line up! Player has max exposure set to 0.')
        if player in self._locked_players:
            raise LineupOptimizerException('This player already in your line up!')
        if self.remaining_budget and player.salary > self.remaining_budget:
            raise LineupOptimizerException('Can\'t add this player to line up! Your team is over budget!')
        if self.remaining_players < 1:
            raise LineupOptimizerException('Can\'t add this player to line up! You already select all %s players!' %
                                           len(self.locked_players))
        max_from_one_team = self._settings.max_from_one_team
        if max_from_one_team:
            from_same_team = len([p for p in self.locked_players if p.team == player.team])
            if from_same_team + 1 > max_from_one_team:
                raise LineupOptimizerException('You can\'t set more than %s players from one team.' %
                                               max_from_one_team)
        position_for_player = None
        if position:
            if position not in self.available_positions:
                raise LineupOptimizerException(
                    'Position %s doesn\'t exist. Available positions are %s' %
                    (position, ','.join(self.available_positions))
                )
            for remaining_position in self._remaining_positions:
                if position:
                    if position == remaining_position.name:
                        position_for_player = remaining_position
                        break
            else:
                raise LineupOptimizerException('Position %s is filled.' % position)
            if not list_intersection(player.positions, position_for_player.positions):
                raise LineupOptimizerException('Player can\'t be set to position %s' % position)
        try:
            link_players_with_positions({player, *self._locked_players}, self._settings.positions)
        except LineupOptimizerException:
            raise LineupOptimizerException('You\'re already select all %s\'s' % '/'.join(player.positions))
        if position_for_player:
            self._remaining_positions.remove(position_for_player)
        self._locked_players[player] = position_for_player

    def unlock_player(self, player: DirtyPlayer):
        player = self._clean_player(player)
        if player not in self._locked_players:
            raise LineupOptimizerException('Player is not locked')
        position = self._locked_players[player]
        if position:
            self._remaining_positions.insert(0, position)
        del self._locked_players[player]

    def get_players(self, *players_and_groups: Union[DirtyPlayer, PlayerFilter]) -> List[Player]:
        result = []
        players, filters = [], []
        for item in players_and_groups:
            if isinstance(item, BaseFilter):
                filters.append(item)
            else:
                players.append(item)
        allowed_players = None
        if filters:
            allowed_players = {
                player for player in self.all_players if
                all(player_filter.filter(player, False) for player_filter in filters)
            }
        if players:
            for player in players:
                cleaned_player = self._clean_player(player, allowed_players=allowed_players)
                if cleaned_player:
                    result.append(cleaned_player)
        return result

    def add_filters(self, *filters: BaseFilter):
        self._player_filters.extend(filters)

    def _clean_player(self, player: DirtyPlayer, allowed_players: Optional[Set[Player]] = None) -> Player:
        if not isinstance(player, Player):
            cleaned_player = self.get_player_by_id(player) or self.get_player_by_name(
                player, allowed_players=allowed_players)
            if not cleaned_player:
                raise LineupOptimizerException('%s is not found' % player)
            return cleaned_player
        return player



from pydfs_lineup_optimizer.sites.yahoo.importer import *
from pydfs_lineup_optimizer.sites.yahoo.settings import *


__all__ = [
    'YahooCSVImporter', 'YahooBasketballSettings', 'YahooFootballSettings', 'YahooHockeySettings',
    'YahooBaseballSettings', 'YahooGolfSettings', 'YahooSoccerSettings',
]



import csv
from typing import Dict, Tuple
from pydfs_lineup_optimizer.exceptions import LineupOptimizerIncorrectCSV
from pydfs_lineup_optimizer.lineup_importer import CSVImporter
from pydfs_lineup_optimizer.player import Player, GameInfo


class YahooCSVImporter(CSVImporter):  # pragma: nocover
    @staticmethod
    def _process_cell(val: str) -> str:
        return val.strip()

    def import_players(self):
        games = {}  # type: Dict[Tuple[str, str], GameInfo]
        players = []
        with open(self.filename, 'r') as csvfile:
            csv_data = csv.DictReader(csvfile, skipinitialspace=True)
            for row in csv_data:
                player_id = self._process_cell(row['ID'])
                if not player_id:
                    continue
                try:
                    away_team, home_team = self._process_cell(row.get('Game', '')).split('@')
                    game_info = games.get((home_team, away_team), None)
                    if not game_info:
                        game_info = GameInfo(home_team, away_team, None)
                        games[(home_team, away_team)] = game_info
                except ValueError:
                    game_info = None
                try:
                    player = Player(
                        player_id,
                        self._process_cell(row['First Name']),
                        self._process_cell(row['Last Name']),
                        self._process_cell(row['Position']).split('/'),
                        self._process_cell(row['Team']),
                        float(self._process_cell(row['Salary'])),
                        float(self._process_cell(row['FPPG'])),
                        is_injured=True if self._process_cell(row['Injury Status']) else False,
                        game_info=game_info,
                        **self.get_player_extra(row)
                    )
                except KeyError:
                    raise LineupOptimizerIncorrectCSV
                players.append(player)
        return players



from collections import defaultdict
from typing import Type, DefaultDict, Dict
from pydfs_lineup_optimizer.settings import BaseSettings


class SitesRegistry:
    SETTINGS_MAPPING = defaultdict(dict)  # type: DefaultDict[str, Dict[str, Type[BaseSettings]]]

    @classmethod
    def register_settings(cls, settings_cls: Type[BaseSettings]) -> Type[BaseSettings]:
        cls.SETTINGS_MAPPING[settings_cls.site][settings_cls.sport] = settings_cls
        return settings_cls

    @classmethod
    def get_settings(cls, site: str, sport: str) -> Type[BaseSettings]:
        try:
            return cls.SETTINGS_MAPPING[site][sport]
        except KeyError:
            raise NotImplementedError



from typing import Dict
from random import getrandbits, uniform
from pydfs_lineup_optimizer.player import Player
from pydfs_lineup_optimizer.lineup import Lineup


class BaseFantasyPointsStrategy:
    def get_player_fantasy_points(self, player: Player) -> float:
        raise NotImplementedError

    def set_previous_lineup(self, lineup: Lineup):
        pass


class StandardFantasyPointsStrategy(BaseFantasyPointsStrategy):
    def get_player_fantasy_points(self, player: Player) -> float:
        return player.fppg

    def set_previous_lineup(self, lineup: Lineup):
        pass


class RandomFantasyPointsStrategy(BaseFantasyPointsStrategy):
    def __init__(self, min_deviation: float = 0.0, max_deviation: float = 0.12):
        self.min_deviation = min_deviation
        self.max_deviation = max_deviation

    def get_player_fantasy_points(self, player: Player) -> float:
        if player.fppg_floor is not None and player.fppg_ceil is not None:
            return uniform(player.fppg_floor, player.fppg_ceil)
        multiplier = uniform(
            player.min_deviation if player.min_deviation is not None else self.min_deviation,
            player.max_deviation if player.max_deviation is not None else self.max_deviation
        )
        return player.fppg * (1 + (-1 if bool(getrandbits(1)) else 1) * multiplier)


class ProgressiveFantasyPointsStrategy(BaseFantasyPointsStrategy):
    def __init__(self, scale: float = 0.01):
        self.scale = scale
        self.player_multipliers = {}  # type: Dict[Player, float]

    def get_player_fantasy_points(self, player: Player) -> float:
        if player not in self.player_multipliers:
            self.player_multipliers[player] = 0
        return player.fppg * (1 + self.player_multipliers[player])

    def set_previous_lineup(self, lineup: Lineup):
        lineup_players = set(lineup)
        for player in self.player_multipliers:
            if player in lineup_players:
                self.player_multipliers[player] = 0
            else:
                scale = player.progressive_scale if player.progressive_scale is not None else self.scale
                self.player_multipliers[player] += scale



from copy import copy
from typing import cast, Optional, List, Union, Iterable
from pydfs_lineup_optimizer.solvers.base import Solver
from pydfs_lineup_optimizer.solvers.constants import SolverSign
from pydfs_lineup_optimizer.solvers.exceptions import SolverException, SolverInfeasibleSolutionException
try:
    from mip import Model, maximize, xsum, Var
    from mip.constants import MAXIMIZE, BINARY, INTEGER, OptimizationStatus
except ImportError:
    raise ImportError('You should install mip library before using this backend')


class MIPVariable:
    def __init__(
            self,
            name: str,
            min_value: Optional[int] = None,
            max_value: Optional[int] = None,
            multiplier: Optional[int] = None
    ):
        self.name = name
        self.min_value = min_value
        self.max_value = max_value
        self.multiplier = multiplier
        self.__cache = None

    def setup(self, solver: Model):
        if any([self.min_value, self.max_value]):
            var = solver.add_var(name=self.name, lb=self.min_value, ub=self.max_value, var_type=INTEGER)
        else:
            var = solver.add_var(name=self.name, var_type=BINARY)
        self.__cache = var

    def get_var(self, solver: Model) -> Var:
        if self.__cache:
            var = self.__cache
        else:
            var = solver.var_by_name(self.name)
            self.__cache = var
        if self.multiplier:
            return self.multiplier * var
        return var

    def __mul__(self, other: int) -> 'MIPVariable':
        return MIPVariable(self.name, self.min_value, self.max_value, other)

    def __rmul__(self, other: int) -> 'MIPVariable':
        return self * other


class MIPConstraint:
    def __init__(
            self,
            variables: Iterable[MIPVariable],
            coefficients: Optional[Iterable[float]],
            sign: str,
            rhs: Union[float, MIPVariable],
            name: Optional[str] = None
    ):
        self.variables = variables
        self.coefficients = coefficients
        self.sign = sign
        self.rhs = rhs
        self.name = name

    def __str__(self):
        return f'{[var.name for var in self.variables]} {self.sign} {self.rhs}'

    def setup(self, solver):
        variables = self.variables
        coefficients = self.coefficients
        sign = self.sign
        rhs = self.rhs
        name = self.name
        if coefficients:
            lhs = xsum([variable.get_var(solver) * coefficient for variable, coefficient in zip(variables, coefficients)])
        else:
            lhs = xsum([var.get_var(solver) for var in variables])
        if isinstance(rhs, MIPVariable):
            rhs = rhs.get_var(solver)
        if sign == SolverSign.EQ:
            solver += (lhs == rhs, name) if name else lhs == rhs
        elif sign == SolverSign.NOT_EQ:
            solver += (lhs != rhs, name) if name else lhs != rhs
        elif sign == SolverSign.GTE:
            solver += (lhs >= rhs, name) if name else lhs >= rhs
        elif sign == SolverSign.LTE:
            solver += (lhs <= rhs, name) if name else lhs <= rhs
        else:
            raise SolverException('Incorrect constraint sign')


class MIPObjective:
    def __init__(self, variables: List[MIPVariable], coefficients: List[float]):
        self.variables = variables
        self.coefficients = coefficients

    def setup(self, solver: Model):
        solver.objective = maximize(xsum(
            variable.get_var(solver) * coefficient for variable, coefficient in zip(self.variables, self.coefficients)))


class MIPSolver(Solver):
    def __init__(self):
        self.model = None  # type: Optional[Model]
        self._vars = {}
        self._constraints = []
        self._objective = None

    def setup_solver(self) -> None:
        self.model = Model(name='pydfs_lineup_optimizer', sense=MAXIMIZE)
        self.model.solver.set_verbose(0)

    def add_variable(self, name, min_value=None, max_value=None):
        var = MIPVariable(name, min_value, max_value)
        if name in self._vars:
            raise ValueError(name)
        self._vars[name] = var
        return var

    def set_objective(self, variables, coefficients):
        self._objective = MIPObjective(variables, coefficients)

    def add_constraint(self, variables, coefficients, sign, rhs, name=None):
        self._constraints.append(MIPConstraint(
            variables,
            coefficients,
            sign,
            rhs,
            name
        ))

    def copy(self):
        new_solver = type(self)()
        new_solver.setup_solver()
        new_solver._vars = copy(self._vars)
        new_solver._constraints = copy(self._constraints)
        new_solver._objective = self._objective
        return new_solver

    def solve(self):
        model = cast(Model, self.model)
        for var in self._vars.values():
            var.setup(model)
        for constraint in self._constraints:
            constraint.setup(model)
        cast(MIPObjective, self._objective).setup(model)
        status = model.optimize()
        if status not in (OptimizationStatus.OPTIMAL, OptimizationStatus.FEASIBLE):
            raise SolverInfeasibleSolutionException([])
        result = []
        for variable in model.vars:  # type: ignore
            val = variable.x
            if val is not None and round(val) >= 1.0:
                result.append(self._vars[variable.name])
        return result



from pulp import LpProblem, LpMaximize, LpVariable, lpSum, LpStatusOptimal, LpBinary, LpInteger, PULP_CBC_CMD
from pydfs_lineup_optimizer.solvers.base import Solver
from pydfs_lineup_optimizer.solvers.constants import SolverSign
from pydfs_lineup_optimizer.solvers.exceptions import SolverException, SolverInfeasibleSolutionException


class PuLPSolver(Solver):
    LP_SOLVER = PULP_CBC_CMD(msg=False)

    def __init__(self):
        self.prob = LpProblem('pydfs_lineup_optimizer', LpMaximize)

    def setup_solver(self):
        pass

    def add_variable(self, name, min_value=None, max_value=None):
        name = name.replace(' ', '_')
        if any([min_value, max_value]):
            return LpVariable(name, lowBound=min_value, upBound=max_value, cat=LpInteger)
        return LpVariable(name, cat=LpBinary)

    def set_objective(self, variables, coefficients):
        self.prob += lpSum([variable * coefficient for variable, coefficient in zip(variables, coefficients)])

    def add_constraint(self, variables, coefficients, sign, rhs, name=None):
        if coefficients:
            lhs = [variable * coefficient for variable, coefficient in zip(variables, coefficients)]
        else:
            lhs = variables
        if sign == SolverSign.EQ:
            self.prob += lpSum(lhs) == rhs, name
        elif sign == SolverSign.NOT_EQ:
            self.prob += lpSum(lhs) != rhs, name
        elif sign == SolverSign.GTE:
            self.prob += lpSum(lhs) >= rhs, name
        elif sign == SolverSign.LTE:
            self.prob += lpSum(lhs) <= rhs, name
        else:
            raise SolverException('Incorrect constraint sign')

    def copy(self):
        new_solver = type(self)()
        new_solver.prob = self.prob.copy()
        return new_solver

    def solve(self):
        self.prob.solve(self.LP_SOLVER)
        if self.prob.status == LpStatusOptimal:
            result = []
            for variable in self.prob.variables():
                val = variable.value()
                if val is not None and round(val) >= 1.0:
                    result.append(variable)
            return result
        else:
            invalid_constraints = [(name or str(constraint)) for name, constraint in
                                   self.prob.constraints.items() if not constraint.valid()]
            raise SolverInfeasibleSolutionException(invalid_constraints)



from collections import OrderedDict
from itertools import chain
from math import ceil
from typing import FrozenSet, Type, Generator, Tuple, Optional, List, Dict, Set, Iterable
from pydfs_lineup_optimizer.lineup import Lineup
from pydfs_lineup_optimizer.solvers import Solver, SolverInfeasibleSolutionException
from pydfs_lineup_optimizer.exceptions import LineupOptimizerException, LineupOptimizerIncorrectTeamName, \
    LineupOptimizerIncorrectPositionName, GenerateLineupException
from pydfs_lineup_optimizer.lineup_importer import CSVImporter
from pydfs_lineup_optimizer.settings import BaseSettings
from pydfs_lineup_optimizer.player import Player, LineupPlayer, GameInfo
from pydfs_lineup_optimizer.utils import ratio, link_players_with_positions, get_remaining_positions, \
    show_deprecation_warning
from pydfs_lineup_optimizer.rules import *
from pydfs_lineup_optimizer.stacks import BaseGroup, BaseStack, Stack
from pydfs_lineup_optimizer.context import OptimizationContext
from pydfs_lineup_optimizer.statistics import Statistic
from pydfs_lineup_optimizer.exposure_strategy import BaseExposureStrategy, TotalExposureStrategy
from pydfs_lineup_optimizer.fantasy_points_strategy import BaseFantasyPointsStrategy, StandardFantasyPointsStrategy, \
    RandomFantasyPointsStrategy
from pydfs_lineup_optimizer.solvers import get_default_solver
from pydfs_lineup_optimizer.player_pool import PlayerPool


BASE_RULES = {
    TotalPlayersRule, LineupBudgetRule, PositionsRule, MaxFromOneTeamRule, LockedPlayersRule,
    UniquePlayerRule, UniqueLineupRule, TotalTeamsRule, GenericStacksRule, MinExposureRule, MinGamesRule,
}


class LineupOptimizer:
    def __init__(self, settings: Type[BaseSettings], solver: Type[Solver] = get_default_solver()):
        self._settings = settings()
        self.player_pool = PlayerPool(self._settings)
        self._csv_importer = None  # type: Optional[Type[CSVImporter]]
        self._rules = BASE_RULES.copy()  # type: Set[Type[OptimizerRule]]
        self._min_deviation = 0.0
        self._max_deviation = 0.12
        self.players_from_one_team = {}  # type: Dict[str, int]
        self.players_with_same_position = {}  # type: Dict[str, int]
        self.min_salary_cap = None  # type: Optional[float]
        self.max_repeating_players = None  # type: Optional[int]
        self._solver_class = solver
        self.max_projected_ownership = None  # type: Optional[float]
        self.min_projected_ownership = None  # type: Optional[float]
        self.opposing_teams_position_restriction = None  # type: Optional[Tuple[List[str], List[str]]]
        self.spacing_positions = None  # type: Optional[List[str]]
        self.spacing = None  # type: Optional[int]
        self.default_team_exposure = None  # type: Optional[float]
        self.teams_exposures = None  # type: Optional[Dict[str, float]]
        self.teams_exposure_strategy = None  # type: Optional[Type[BaseExposureStrategy]]
        self.team_stacks_for_positions = None  # type: Optional[List[str]]
        self.same_team_restrict_positions = None  # type: Optional[Tuple[Tuple[str, str], ...]]
        self.opposing_team_force_positions = None  # type: Optional[Tuple[Tuple[str, str], ...]]
        self.opposing_teams_max_allowed = 0
        self.min_teams = None  # type: Optional[int]
        self.max_teams = None  # type: Optional[int]
        self.stacks = []  # type: List[BaseStack]
        self.min_starters = None  # type: Optional[int]
        self.last_context = None  # type: Optional[OptimizationContext]
        self.fantasy_points_strategy = StandardFantasyPointsStrategy()  # type: BaseFantasyPointsStrategy

    @property
    def budget(self) -> Optional[float]:
        return self._settings.budget

    @property
    def total_players(self) -> int:
        show_deprecation_warning('total_players will be removed in version 3.7, use player_pool.total_players instead')
        return self.player_pool.total_players

    @property
    def remaining_budget(self) -> Optional[float]:
        show_deprecation_warning('remaining_budget will be removed in version 3.7, '
                                 'use player_pool.remaining_budget instead')
        return self.player_pool.remaining_budget

    @property
    def remaining_players(self) -> int:
        show_deprecation_warning('remaining_players will be removed in version 3.7, '
                                 'use player_pool.remaining_players instead')
        return self.player_pool.remaining_players

    @property
    def max_from_one_team(self) -> Optional[int]:
        return self._settings.max_from_one_team

    @property
    def available_teams(self) -> FrozenSet[str]:
        show_deprecation_warning('available_teams will be removed in version 3.7, '
                                 'use player_pool.available_teams instead')
        return self.player_pool.available_teams

    @property
    def available_positions(self) -> FrozenSet[str]:
        show_deprecation_warning('available_positions will be removed in version 3.7, '
                                 'use player_pool.available_positions instead')
        return self.player_pool.available_positions

    @property
    def removed_players(self) -> Set[Player]:
        show_deprecation_warning('removed_players will be removed in version 3.7, '
                                 'use player_pool.removed_players instead')
        return self.player_pool.removed_players

    @property
    def players(self) -> List[Player]:
        show_deprecation_warning('players will be removed in version 3.7, '
                                 'use player_pool.filtered_players instead')
        return self.player_pool.filtered_players

    @property
    def locked_players(self) -> List[Player]:
        show_deprecation_warning('locked_players will be removed in version 3.7, use player_pool.locked_players instead')
        return self.player_pool.locked_players

    @property
    def settings(self) -> BaseSettings:
        return self._settings

    @property
    def games(self) -> FrozenSet[GameInfo]:
        show_deprecation_warning('games will be removed in version 3.7, use player_pool.games instead')
        return self.player_pool.games

    def reset_lineup(self):
        show_deprecation_warning('reset_lineup will be removed in version 3.7, use player_pool.reset_locked instead')
        self.player_pool.reset_locked()

    def set_fantasy_points_strategy(self, strategy: BaseFantasyPointsStrategy):
        self.fantasy_points_strategy = strategy

    def set_deviation(self, min_deviation: float, max_deviation: float):
        """
        Set deviation ranges for randomness mode
        """
        show_deprecation_warning('set_deviation method is deprecated and will be removed in 3.6, '
                                 'set deviation in  RandomFantasyPointsStrategy instead')
        self._min_deviation = min_deviation
        self._max_deviation = max_deviation

    def get_deviation(self) -> Tuple[float, float]:
        return self._min_deviation, self._max_deviation

    def set_csv_importer(self, csv_importer: Type[CSVImporter]):
        self._csv_importer = csv_importer

    def set_min_salary_cap(self, min_salary: float):
        if self.budget and min_salary > self.budget:
            raise LineupOptimizerException('Min salary greater than max budget')
        self.add_new_rule(MinSalaryCapRule)
        self.min_salary_cap = min_salary

    def load_players_from_csv(self, filename: str):
        """
        Load player list from CSV file with passed filename.
        """
        csv_importer = self._csv_importer
        if not csv_importer:
            csv_importer = self._settings.csv_importer
        self.player_pool.extend_players(csv_importer(filename).import_players())

    def load_lineups_from_csv(self, filename: str) -> List[Lineup]:
        csv_importer = self._csv_importer
        if not csv_importer:
            csv_importer = self._settings.csv_importer
        return csv_importer(filename).import_lineups(self.player_pool.all_players)

    def load_players(self, players: List[Player]):
        show_deprecation_warning('load_players will be removed in version 3.7, use player_pool.load_players instead')
        self.player_pool.load_players(players)

    def extend_players(self, players: List[Player]):
        show_deprecation_warning('extend_players will be removed in version 3.7, use player_pool.extend_players instead')
        self.player_pool.extend_players(players)

    def add_new_rule(self, rule: Type[OptimizerRule]):
        self._rules.add(rule)

    def remove_rule(self, rule: Type[OptimizerRule], silent: bool = True):
        try:
            self._rules.remove(rule)
        except KeyError:
            if not silent:
                raise LineupOptimizerException('Rule isn\'t added!')

    def remove_player(self, player: Player):
        show_deprecation_warning('remove_player will be removed in version 3.7, use player_pool.remove_player instead')
        self.player_pool.remove_player(player)

    def restore_player(self, player: Player):
        show_deprecation_warning('restore_player will be removed in version 3.7, use player_pool.restore_player instead')
        self.player_pool.restore_player(player)

    def find_players(self, name: str) -> List[Player]:
        """
        Return list of players with similar name.
        """
        show_deprecation_warning('find_players will be removed in version 3.7')
        player_pool = self.player_pool
        possibilities = [(player, ratio(name, player.full_name)) for player in player_pool.all_players]
        filtered_possibilities = filter(lambda pos: pos[1] >= player_pool.search_threshold, possibilities)
        players = sorted(filtered_possibilities, key=lambda pos: -pos[1])
        return list(map(lambda p: p[0], players))

    def get_player_by_name(self, name: str) -> Optional[Player]:
        show_deprecation_warning('get_player_by_name will be removed in version 3.7, '
                                 'use player_pool.get_player_by_name instead')
        return self.player_pool.get_player_by_name(name)

    def get_player_by_id(self, player_id: str) -> Optional[Player]:
        show_deprecation_warning('get_player_by_id will be removed in version 3.7, '
                                 'use player_pool.get_player_by_id instead')
        return self.player_pool.get_player_by_id(player_id)

    def add_player_to_lineup(self, player: Player):
        show_deprecation_warning('add_player_to_lineup will be removed in version 3.7, '
                                 'use player_pool.lock_player instead')
        self.player_pool.lock_player(player)

    def remove_player_from_lineup(self, player: Player):
        show_deprecation_warning('remove_player_from_lineup will be removed in version 3.7, '
                                 'use player_pool.unlock_player instead')
        self.player_pool.unlock_player(player)

    def set_players_from_one_team(self, teams: Optional[Dict[str, int]] = None):
        if teams is not None:
            teams = {team: num_of_players for team, num_of_players in teams.items()}
            for team, num_of_players in teams.items():
                self._check_team_constraint(team, num_of_players)
            self.add_new_rule(TeamMatesRule)
        else:
            self.remove_rule(TeamMatesRule)
        self.players_from_one_team = teams or {}

    def set_players_with_same_position(self, positions: Dict[str, int]):
        positions = positions or {}
        positions = {position.upper(): num_of_players for position, num_of_players in positions.items()}
        for pos, val in positions.items():
            self._check_position_constraint(pos)
        self.players_with_same_position = positions

    def set_max_repeating_players(self, max_repeating_players: int):
        if max_repeating_players >= self.player_pool.total_players:
            raise LineupOptimizerException('Maximum repeating players should be smaller than %d' %
                                           self.player_pool.total_players)
        elif max_repeating_players < 1:
            raise LineupOptimizerException('Maximum repeating players should be 1 or greater')
        self.max_repeating_players = max_repeating_players

    def set_projected_ownership(
            self,
            min_projected_ownership: Optional[float] = None,
            max_projected_ownership: Optional[float] = None
    ):
        if min_projected_ownership and max_projected_ownership and min_projected_ownership >= max_projected_ownership:
            raise LineupOptimizerException('Max projected ownership should be greater than min projected ownership')
        self.max_projected_ownership = max_projected_ownership / 100 if \
            max_projected_ownership and max_projected_ownership > 1 else max_projected_ownership
        self.min_projected_ownership = min_projected_ownership / 100 if \
            min_projected_ownership and min_projected_ownership > 1 else min_projected_ownership
        if max_projected_ownership or min_projected_ownership:
            self.add_new_rule(ProjectedOwnershipRule)
        else:
            self.remove_rule(ProjectedOwnershipRule)

    def restrict_positions_for_opposing_team(
            self,
            first_team_positions: List[str],
            second_team_positions: List[str],
            max_allowed: int = 0,
    ):
        if not self.player_pool.games:
            raise LineupOptimizerException('Game Info isn\'t specified for players')
        self.opposing_teams_position_restriction = (first_team_positions, second_team_positions)
        self.opposing_teams_max_allowed = max_allowed
        self.add_new_rule(RestrictPositionsForOpposingTeam)

    def restrict_positions_for_same_team(self, *restrict_positions: Tuple[str, str]):
        if not all(len(positions) == 2 for positions in restrict_positions):
            raise LineupOptimizerException('Exactly 2 positions must be specified in restrict positions')
        for position in set(chain.from_iterable(restrict_positions)):
            self._check_position_constraint(position)
        self.same_team_restrict_positions = restrict_positions
        self.add_new_rule(RestrictPositionsForSameTeamRule)

    def force_positions_for_opposing_team(self, *force_positions: Tuple[str, str]):
        if not all(len(positions) == 2 for positions in force_positions):
            raise LineupOptimizerException('Exactly 2 positions must be specified in force positions')
        for position in set(chain.from_iterable(force_positions)):
            self._check_position_constraint(position)
        self.opposing_team_force_positions = force_positions
        self.add_new_rule(ForcePositionsForOpposingTeamRule)

    def set_spacing_for_positions(self, positions: List[str], spacing: int):
        if spacing < 1:
            raise LineupOptimizerException('Spacing must be 1 or greater')
        available_positions = self.player_pool.available_positions
        if any(position not in available_positions for position in positions):
            raise LineupOptimizerException('Incorrect positions. Choices are: %s' % available_positions)
        self.spacing_positions = positions
        self.spacing = spacing
        self.add_new_rule(RosterSpacingRule)

    def set_total_teams(self, total_teams: Optional[int] = None, min_teams: Optional[int] = None,
                        max_teams: Optional[int] = None):
        if not any((total_teams, min_teams, max_teams)):
            raise LineupOptimizerException('At least one parameter should be passed')
        if total_teams:
            min_teams = total_teams
            max_teams = total_teams
            show_deprecation_warning('total_teams parameter is replaced with min_teams/max_teams parameters and '
                                     'will be removed in version 3.7')
        available_teams = len(self.player_pool.available_teams)
        if min_teams is not None and max_teams is not None and min_teams > max_teams:
            raise LineupOptimizerException('min_teams can\'t be greater than max_teams')
        if (min_teams is not None and min_teams > available_teams) or \
                (max_teams is not None and max_teams > available_teams):
            raise LineupOptimizerException('Only %d teams available' % available_teams)
        total_players = self.settings.get_total_players()
        settings_min_teams = self.settings.min_teams
        max_from_one_team = self.settings.max_from_one_team
        if not settings_min_teams and max_from_one_team:
            settings_min_teams = ceil(total_players / max_from_one_team)
        if settings_min_teams and ((
            max_teams is not None and max_teams < settings_min_teams
        ) or (
            min_teams is not None and min_teams < settings_min_teams
        )):
            raise LineupOptimizerException('Minimum number of teams is %d' % settings_min_teams)
        if (min_teams is not None and min_teams > total_players) or \
                (max_teams is not None and max_teams > total_players):
            raise LineupOptimizerException('Maximum number of teams is %d' % total_players)
        self.min_teams = min_teams
        self.max_teams = max_teams

    def add_players_group(self, group: BaseGroup) -> None:
        stack = Stack(groups=[group])
        self.stacks.append(stack)

    def add_stack(self, stack: BaseStack) -> None:
        stack.validate(self)
        self.stacks.append(stack)

    def reset_stacks(self) -> None:
        self.stacks = []

    def set_min_starters(self, min_starters: int) -> None:
        if min_starters > self.settings.get_total_players():
            raise LineupOptimizerException('Num of starters can\'t be greater than max players')
        if min_starters > len([p for p in self.player_pool.filtered_players if p.is_confirmed_starter]):
            raise LineupOptimizerException('Num of starters can\'t be greater than max starters')
        self.min_starters = min_starters
        self.add_new_rule(MinStartersRule)

    def set_teams_max_exposures(
            self,
            default_max_exposure: Optional[float] = None,
            exposures_by_team: Optional[Dict[str, float]] = None,
            exposure_strategy: Optional[Type[BaseExposureStrategy]] = TotalExposureStrategy,
    ):
        if exposures_by_team:
            for team in exposures_by_team:
                if team not in self.player_pool.available_teams:
                    raise LineupOptimizerException('Unknown team: %s' % team)
        self.default_team_exposure = default_max_exposure
        self.teams_exposures = exposures_by_team
        self.teams_exposure_strategy = exposure_strategy
        if default_max_exposure or exposures_by_team:
            self.add_new_rule(TeamsExposureRule)
        else:
            self.remove_rule(TeamsExposureRule)

    def optimize(
            self,
            n: int,
            max_exposure: Optional[float] = None,
            randomness: bool = False,
            with_injured: Optional[bool] = None,
            exposure_strategy: Type[BaseExposureStrategy] = TotalExposureStrategy,
            exclude_lineups: Optional[Iterable[Lineup]] = None,
    ) -> Generator[Lineup, None, None]:
        if with_injured is not None:
            show_deprecation_warning('with_injured parameter is deprecated, use player_pool.with_injured instead')
            self.player_pool.with_injured = with_injured
        players = self.player_pool.filtered_players
        context = OptimizationContext(
            total_lineups=n,
            players=players,
            max_exposure=max_exposure,
            randomness=randomness,
            exposure_strategy=exposure_strategy,
            exclude_lineups=exclude_lineups or [],
        )
        rules = self._rules.copy()
        rules.update(self.settings.extra_rules)
        if randomness:
            show_deprecation_warning('Randomness parameter is deprecated and will be removed in 3.6, '
                                     'use set_fantasy_points_strategy instead')
            self.set_fantasy_points_strategy(RandomFantasyPointsStrategy(self._min_deviation, self._max_deviation))
        rules.add(Objective)
        base_solver = self._solver_class()
        base_solver.setup_solver()
        players_dict = OrderedDict(
            [(player, base_solver.add_variable(base_solver.build_player_var_name(player, str(i))))
             for i, player in enumerate(players)])
        variables_dict = {v: k for k, v in players_dict.items()}
        constraints = [constraint(self, players_dict, context) for constraint in rules]
        for constraint in constraints:
            constraint.apply(base_solver)
        previous_lineup = None
        for _ in range(n):
            solver = base_solver.copy()  # type: Solver
            for constraint in constraints:
                constraint.apply_for_iteration(solver, previous_lineup)
            try:
                solved_variables = solver.solve()
                lineup_players = []
                variables_names = []
                for solved_variable in solved_variables:
                    player = variables_dict.get(solved_variable)
                    if player:
                        lineup_players.append(player)
                    variables_names.append(solved_variable.name)
                lineup = self._build_lineup(lineup_players, context)
                previous_lineup = lineup
                context.add_lineup(lineup)
                yield lineup
                total_players = self.player_pool.total_players
                if total_players and len(self.player_pool.locked_players) == total_players:
                    return
                for constraint in constraints:
                    constraint.post_optimize(variables_names)
            except SolverInfeasibleSolutionException as solver_exception:
                raise GenerateLineupException(solver_exception.get_user_defined_constraints())
        self.last_context = context

    def optimize_lineups(
            self,
            lineups: List[Lineup],
            max_exposure: Optional[float] = None,
            randomness: bool = False,
            with_injured: bool = None,
            exposure_strategy: Type[BaseExposureStrategy] = TotalExposureStrategy
    ):
        if with_injured is not None:
            show_deprecation_warning('with_injured parameter is deprecated, use player_pool.with_injured instead')
            self.player_pool.with_injured = with_injured
        players = self.player_pool.filtered_players
        context = OptimizationContext(
            total_lineups=len(lineups),
            players=players,
            existed_lineups=lineups,
            max_exposure=max_exposure,
            randomness=randomness,
            exposure_strategy=exposure_strategy
        )
        rules = self._rules.copy()
        rules.update(self.settings.extra_rules)
        if randomness:
            show_deprecation_warning('Randomness parameter is deprecated and will be removed in 3.6, '
                                     'use set_fantasy_points_strategy instead')
            self.set_fantasy_points_strategy(RandomFantasyPointsStrategy(self._min_deviation, self._max_deviation))
        rules.add(Objective)
        rules.add(LateSwapRule)
        rules.remove(PositionsRule)
        base_solver = self._solver_class()
        base_solver.setup_solver()
        players_dict = OrderedDict(
            [(player, base_solver.add_variable(base_solver.build_player_var_name(player, str(i))))
             for i, player in enumerate(players)])
        variables_dict = {v: k for k, v in players_dict.items()}
        constraints = [constraint(self, players_dict, context) for constraint in rules]
        for constraint in constraints:
            constraint.apply(base_solver)
        previous_lineup = None
        for lineup in lineups:
            if len(lineup.get_unswappable_players()) == self.total_players:
                yield lineup
                continue
            solver = base_solver.copy()  # type: Solver
            for constraint in constraints:
                constraint.apply_for_iteration(solver, previous_lineup)
            try:
                solved_variables = solver.solve()
                unswappable_players = lineup.get_unswappable_players()
                lineup_players = []
                variables_names = []
                for solved_variable in solved_variables:
                    player = variables_dict.get(solved_variable)
                    if player:
                        lineup_players.append(player)
                    variables_names.append(solved_variable.name)
                generated_lineup = self._build_lineup(lineup_players, context, unswappable_players)
                previous_lineup = generated_lineup
                context.add_lineup(generated_lineup)
                yield generated_lineup
                for constraint in constraints:
                    constraint.post_optimize(variables_names)
            except SolverInfeasibleSolutionException as solver_exception:
                raise GenerateLineupException(solver_exception.get_user_defined_constraints())
        self.last_context = context

    def print_statistic(self, with_excluded: bool = True) -> None:
        if self.last_context is None:
            raise LineupOptimizerException('You should generate lineups before printing statistic')
        Statistic(self, with_excluded).print_report()

    def export(self, filename: str, with_excluded: bool = True) -> None:
        if self.last_context is None:
            raise LineupOptimizerException('You should generate lineups before printing statistic')
        self.settings.csv_exporter(self.last_context.get_lineups(with_excluded)).export(filename)

    def _build_lineup(
            self,
            players: List[Player],
            context: OptimizationContext,
            unswappable_players: Optional[List[LineupPlayer]] = None,
    ) -> Lineup:
        lineup = []
        positions = self._settings.positions[:]
        if not positions:
            for player in sorted(players, key=lambda p: p.positions[0]):
                lineup.append(LineupPlayer(player, player.positions[0], used_fppg=context.players_used_fppg.get(player)))
            return Lineup(lineup, self._settings.lineup_printer)
        if unswappable_players:
            players = [player for player in players if player not in unswappable_players]
            positions = get_remaining_positions(positions, unswappable_players)
            lineup.extend(unswappable_players)
        for locked_player, locked_position in self.player_pool.locked_players_with_positions.items():
            players.remove(locked_player)
            positions.remove(locked_position)
            lineup.append(LineupPlayer(locked_player, locked_position.name))
        players_with_positions = link_players_with_positions(players, positions)
        for player, position in players_with_positions.items():
            lineup.append(LineupPlayer(player, position.name, used_fppg=context.players_used_fppg.get(player)))
        positions_order = [pos.name for pos in self._settings.positions]
        lineup.sort(key=lambda p: positions_order.index(p.lineup_position))
        return Lineup(lineup, self._settings.lineup_printer)

    def _check_team_constraint(self, team: str, num_of_players: int):
        if team not in self.player_pool.available_teams:
            raise LineupOptimizerIncorrectTeamName('%s is incorrect team name. Choices are [%s]' %
                                                   (team, ','.join(self.player_pool.available_teams)))
        if self.max_from_one_team and num_of_players > self.max_from_one_team:
            raise LineupOptimizerException('You can\'t set more than %s players from one team.' %
                                           self.max_from_one_team)

    def _check_position_constraint(self, position: str):
        if position not in self.player_pool.available_positions:
            raise LineupOptimizerIncorrectPositionName('%s is incorrect position name. Choices are [%s]' %
                                                       (position, ','.join(self.player_pool.available_positions)))



from copy import deepcopy
from typing import Type
from pydfs_lineup_optimizer.sites.fanduel.classic.importer import FanDuelCSVImporter


def build_fanduel_single_game_importer(mvp=True, star=False, pro=False) -> Type[FanDuelCSVImporter]:
    class FanDuelSingleGameCSVImporter(FanDuelCSVImporter):  # pragma: nocover
        def import_players(self):
            players = super().import_players()
            extra_players = []
            for player in players:
                if mvp:
                    mvp_player = deepcopy(player)
                    mvp_player.fppg *= 2
                    mvp_player._original_positions = player.positions
                    mvp_player.positions = ('MVP', )
                    extra_players.append(mvp_player)
                if star:
                    star_player = deepcopy(player)
                    star_player.fppg *= 1.5
                    star_player._original_positions = player.positions
                    star_player.positions = ('STAR', )
                    extra_players.append(star_player)
                if pro:
                    pro_player = deepcopy(player)
                    pro_player.fppg *= 1.2
                    pro_player._original_positions = player.positions
                    pro_player.positions = ('PRO', )
                    extra_players.append(pro_player)
            players.extend(extra_players)
            return players
    return FanDuelSingleGameCSVImporter


class FanDuelSingleGameHockeyCSVImporter(FanDuelCSVImporter):  # pragma: nocover
    def import_players(self):
        players = super().import_players()
        extra_players = []
        for player in players:
            captain_player = deepcopy(player)
            captain_player.fppg *= 1.5
            captain_player._original_positions = player.positions
            captain_player.positions = ('CAPTAIN', )
            extra_players.append(captain_player)
        players.extend(extra_players)
        return players



from typing import List


class LineupOptimizerException(Exception):
    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return self.message


class LineupOptimizerIncorrectTeamName(LineupOptimizerException):
    pass


class LineupOptimizerIncorrectPositionName(LineupOptimizerException):
    pass


class LineupOptimizerIncorrectCSV(LineupOptimizerException):
    def __init__(self, message: str = 'Incorrect csv format!'):
        super(LineupOptimizerIncorrectCSV, self).__init__(message)


class GenerateLineupException(LineupOptimizerException):
    def __init__(self, invalid_constraints: List[str]):
        self.invalid_constraints = invalid_constraints

    def __str__(self):
        msg = 'Can\'t generate lineups.'
        if self.invalid_constraints:
            msg += ' Following constraints are not valid: %s' % ','.join(self.invalid_constraints)
        return msg



from abc import ABCMeta, abstractmethod
from math import floor
from uuid import uuid4
from typing import List, Optional, Tuple, Dict, Union, cast, TYPE_CHECKING
from itertools import chain
from collections import Counter, defaultdict
from pydfs_lineup_optimizer import Player
from pydfs_lineup_optimizer.exceptions import LineupOptimizerException, LineupOptimizerIncorrectTeamName, \
    LineupOptimizerIncorrectPositionName
from pydfs_lineup_optimizer.utils import get_players_grouped_by_teams, list_intersection


if TYPE_CHECKING:  # pragma: no cover
    from pydfs_lineup_optimizer.lineup_optimizer import LineupOptimizer


class OptimizerGroup:
    def __init__(
            self,
            players: List[Player],
            min_from_group: Optional[int] = None,
            max_from_group: Optional[int] = None,
    ):
        self.players = players
        self.min_from_group = min_from_group
        self.max_from_group = max_from_group


class BaseGroup(metaclass=ABCMeta):
    def __init__(
            self,
            max_exposure: Optional[float] = None,
            depends_on: Optional[Player] = None,
            min_from_group: Optional[int] = None,
            max_from_group: Optional[int] = None,
            strict_depend: bool = True,
            parent: Optional['BaseGroup'] = None,
    ):
        self.uuid = uuid4()
        self.max_exposure = max_exposure
        self.parent = parent
        self.depends_on = depends_on
        self.strict_depend = strict_depend
        self.min_from_group = min_from_group
        self.max_from_group = max_from_group

    @abstractmethod
    def get_all_players_groups(self) -> List[OptimizerGroup]:
        pass


class PlayersGroup(BaseGroup):
    def __init__(
            self,
            players: List[Player],
            min_from_group: Optional[int] = None,
            max_from_group: Optional[int] = None,
            max_exposure: Optional[float] = None,
            depends_on: Optional[Player] = None,
            strict_depend: bool = True,
            parent: Optional[BaseGroup] = None,
    ):
        if min_from_group is None and max_from_group is None:
            min_from_group = len(players)
        super().__init__(
            max_exposure=max_exposure,
            parent=parent,
            min_from_group=min_from_group,
            max_from_group=max_from_group,
            depends_on=depends_on,
            strict_depend=strict_depend,
        )
        self.players = list(set(players))

    def get_all_players_groups(self) -> List[OptimizerGroup]:
        return [OptimizerGroup(self.players, self.min_from_group, self.max_from_group)]

    def __str__(self):
        return 'Group: %s' % ','.join(player.full_name for player in self.players)


class NestedPlayersGroup(BaseGroup):
    def __init__(
            self,
            groups: List[PlayersGroup],
            min_from_group: Optional[int] = None,
            max_from_group: Optional[int] = None,
            max_exposure: Optional[float] = None,
            parent: Optional[BaseGroup] = None,
    ):
        if min_from_group is None and max_from_group is None:
            min_from_group = len(groups)
        super().__init__(
            max_exposure=max_exposure,
            parent=parent,
            min_from_group=min_from_group,
            max_from_group=max_from_group,
        )
        self.groups = groups

    def get_all_players_groups(self) -> List[OptimizerGroup]:
        return list(chain.from_iterable(group.get_all_players_groups() for group in self.groups))


class OptimizerStack:
    def __init__(self, groups: List[BaseGroup], can_intersect=False):
        self.groups = groups
        self.uuid = uuid4()
        self.can_intersect = can_intersect

    @property
    def with_exposures(self) -> bool:
        return any(group.max_exposure is not None for group in self.groups)


class BaseStack(metaclass=ABCMeta):
    @abstractmethod
    def validate(self, optimizer: 'LineupOptimizer') -> None:
        pass

    @abstractmethod
    def build_stacks(self, players: List[Player], optimizer: 'LineupOptimizer') -> List[OptimizerStack]:
        pass


class Stack(BaseStack):
    def __init__(self, groups: List[BaseGroup]):
        self.groups = groups

    def build_stacks(self, players: List[Player], optimizer: 'LineupOptimizer') -> List[OptimizerStack]:
        return [OptimizerStack(groups=self.groups)]

    def validate(self, optimizer: 'LineupOptimizer') -> None:
        pass


class TeamStack(BaseStack):
    def __init__(
            self,
            size: int,
            for_teams: Optional[List[str]] = None,
            for_positions: Optional[List[str]] = None,
            spacing: Optional[int] = None,
            max_exposure: Optional[float] = None,
            max_exposure_per_team: Optional[Dict[str, float]] = None,
    ):
        self.size = size
        self.for_teams = for_teams
        self.for_positions = for_positions
        self.spacing = spacing
        self.max_exposure = max_exposure
        self.max_exposure_per_team = max_exposure_per_team or {}

    def build_stacks(self, players: List[Player], optimizer: 'LineupOptimizer') -> List[OptimizerStack]:
        players_by_teams = get_players_grouped_by_teams(
            players, for_teams=self.for_teams, for_positions=self.for_positions)
        groups = []  # type: List[BaseGroup]
        spacing_groups = []  # type: List[BaseGroup]
        for team, players in players_by_teams.items():
            parent_group = PlayersGroup(
                players=players,
                min_from_group=self.size,
                max_exposure=self.max_exposure_per_team.get(team, self.max_exposure)
            )
            groups.append(parent_group)
            if not self.spacing:
                continue
            players_by_roster_positions = defaultdict(list)  # type: Dict[int, List[Player]]
            for player in players:
                if player.roster_order is None:
                    continue
                players_by_roster_positions[player.roster_order].append(player)
            if not players_by_roster_positions:
                continue
            all_allowed_roster_orders = set()
            max_spacing = max(players_by_roster_positions.keys())
            for roster_position in players_by_roster_positions.keys():
                allowed_roster_orders = []
                for i in range(self.spacing):
                    if roster_position + i <= max_spacing:
                        allowed_roster_orders.append(roster_position + i)
                    else:
                        allowed_roster_orders.append(((roster_position + i) % (max_spacing + 1)) + 1)
                all_allowed_roster_orders.add(tuple(sorted(allowed_roster_orders)))
            for roster_orders in all_allowed_roster_orders:
                spacing_groups.append(PlayersGroup(
                    players=list(chain.from_iterable(
                        players for players_spacing, players in players_by_roster_positions.items()
                        if players_spacing in roster_orders
                    )),
                    min_from_group=self.size,
                    parent=parent_group,
                ))
        stacks = [OptimizerStack(groups=groups)]
        if spacing_groups:
            stacks.append(OptimizerStack(groups=spacing_groups, can_intersect=True))
        return stacks

    def validate(self, optimizer: 'LineupOptimizer') -> None:
        settings = optimizer.settings
        max_from_one_team = settings.max_from_one_team or settings.get_total_players()
        if self.size > max_from_one_team:
            raise LineupOptimizerException(
                'Stack size should be smaller than max players from one team (%d)' % max_from_one_team)
        for team in self.for_teams or []:
            if team not in optimizer.available_teams:
                raise LineupOptimizerIncorrectTeamName('%s is incorrect team name.' % team)
        for position in self.for_positions or []:
            if position not in optimizer.player_pool.available_positions:
                raise LineupOptimizerIncorrectPositionName('%s is incorrect position name.' % position)


class PositionsStack(BaseStack):
    def __init__(
            self,
            positions: List[Union[str, Tuple[str, ...]]],
            for_teams: Optional[List[str]] = None,
            max_exposure: Optional[float] = None,
            max_exposure_per_team: Optional[Dict[str, float]] = None,
    ):
        formatted_positions_stacks = tuple(
            [cast(Tuple[str, ...], (position,) if isinstance(position, str) else tuple(position))
             for position in positions]
        )
        self.positions = formatted_positions_stacks
        self.for_teams = for_teams
        self.max_exposure = max_exposure
        self.max_exposure_per_team = max_exposure_per_team or {}

    def build_stacks(self, players: List[Player], optimizer: 'LineupOptimizer') -> List[OptimizerStack]:
        players_by_teams = get_players_grouped_by_teams(players, for_teams=self.for_teams)
        all_positions = tuple(set(chain.from_iterable(self.positions)))
        positions_for_optimizer = Counter(self.positions)
        positions_for_optimizer[all_positions] = len(self.positions)
        all_groups = []  # type: List[BaseGroup]
        for team_name, team_players in players_by_teams.items():
            groups = []
            for positions, total in positions_for_optimizer.items():
                groups.append(PlayersGroup(
                    players=[player for player in team_players if list_intersection(player.positions, positions)],
                    min_from_group=total,
                ))
            nested_group = NestedPlayersGroup(
                groups=groups,
                max_exposure=self.max_exposure_per_team.get(team_name, self.max_exposure),
            )
            all_groups.append(nested_group)
        return [OptimizerStack(groups=all_groups)]

    def validate(self, optimizer: 'LineupOptimizer') -> None:
        if not self.positions or not all(self.positions):
            raise LineupOptimizerException('Positions stack can\'t be empty')
        settings = optimizer.settings
        max_from_one_team = settings.max_from_one_team or settings.get_total_players()
        if len(self.positions) > max_from_one_team:
            raise LineupOptimizerException('You can\'t set more than %s players from one team.' %
                                           max_from_one_team)
        for position in set(chain.from_iterable(self.positions)):
            if position not in optimizer.player_pool.available_positions:
                raise LineupOptimizerIncorrectPositionName('%s is incorrect position name.' % position)
        for team in self.for_teams or []:
            if team not in optimizer.player_pool.available_teams:
                raise LineupOptimizerIncorrectTeamName('%s is incorrect team name.' % team)


class GameStack(BaseStack):
    def __init__(
            self,
            size: int,
            max_exposure: Optional[float] = None,
            min_from_team: int = 1,
    ):
        self.size = size
        self.max_exposure = max_exposure
        self.min_from_team = min_from_team

    def build_stacks(self, players, optimizer):
        players_by_teams = get_players_grouped_by_teams(players)
        all_groups: List[BaseGroup] = []
        for game in optimizer.games:
            groups = [PlayersGroup(
                players=players_by_teams[game.home_team],
                min_from_group=self.min_from_team,
            ), PlayersGroup(
                players=players_by_teams[game.away_team],
                min_from_group=self.min_from_team,
            ), PlayersGroup(
                players=[*players_by_teams[game.home_team], *players_by_teams[game.away_team]],
                min_from_group=self.size,
            )]
            nested_group = NestedPlayersGroup(
                groups=groups,
                max_exposure=self.max_exposure,
            )
            all_groups.append(nested_group)
        return [OptimizerStack(groups=all_groups)]

    def validate(self, optimizer) -> None:
        if floor(self.size / 2) < self.min_from_team:
            raise LineupOptimizerException('min_from_team is greater than stack size')



from uuid import uuid4
from math import ceil
from collections import defaultdict, Counter
from itertools import product, groupby, permutations, chain
from typing import List, Dict, DefaultDict, Set, Tuple, Any, Optional, TYPE_CHECKING
from weakref import proxy
from pydfs_lineup_optimizer.solvers import Solver, SolverSign
from pydfs_lineup_optimizer.utils import list_intersection, get_positions_for_optimizer, get_remaining_positions, \
    get_players_grouped_by_teams
from pydfs_lineup_optimizer.lineup import Lineup
from pydfs_lineup_optimizer.player import Player, GameInfo
from pydfs_lineup_optimizer.context import OptimizationContext


if TYPE_CHECKING:  # pragma: no cover
    from pydfs_lineup_optimizer.lineup_optimizer import LineupOptimizer
    from pydfs_lineup_optimizer.stacks import BaseGroup


__all__ = [
    'OptimizerRule', 'Objective', 'UniqueLineupRule', 'TotalPlayersRule',
    'LineupBudgetRule', 'LockedPlayersRule', 'PositionsRule', 'TeamMatesRule', 'MaxFromOneTeamRule',
    'MinSalaryCapRule',
    'ProjectedOwnershipRule', 'UniquePlayerRule', 'LateSwapRule',
    'RestrictPositionsForOpposingTeam', 'RosterSpacingRule', 'FanduelBaseballRosterRule',
    'TotalTeamsRule', 'FanduelSingleGameMaxQBRule',
    'RestrictPositionsForSameTeamRule', 'ForcePositionsForOpposingTeamRule', 'GenericStacksRule',
    'MinStartersRule', 'MinExposureRule', 'MinGamesRule', 'DraftKingsBaseballRosterRule',
    'DraftKingsTiersRule', 'TeamsExposureRule',
]


class OptimizerRule:
    def __init__(self, optimizer: 'LineupOptimizer', players_dict: Dict[Player, Any], context: OptimizationContext):
        self.optimizer = proxy(optimizer)
        self.context = context
        self.players_dict = players_dict

    @property
    def player_pool(self):
        return self.optimizer.player_pool

    def apply(self, solver: Solver):
        pass

    def apply_for_iteration(self, solver: Solver, result: Optional[Lineup]):
        pass

    def post_optimize(self, solved_variables: List[str]):
        pass


class Objective(OptimizerRule):
    def __init__(self, optimizer, players_dict, context):
        super().__init__(optimizer, players_dict, context)
        self.fantasy_points_strategy = optimizer.fantasy_points_strategy

    def apply_for_iteration(self, solver, result):
        if result:
            self.fantasy_points_strategy.set_previous_lineup(result)
        variables = []
        coefficients = []
        get_points = self.fantasy_points_strategy.get_player_fantasy_points
        for player, variable in self.players_dict.items():
            fantasy_points = get_points(player)
            variables.append(variable)
            coefficients.append(fantasy_points)
            self.context.players_used_fppg[player] = fantasy_points
        solver.set_objective(variables, coefficients)


class UniqueLineupRule(OptimizerRule):
    def __init__(self, optimizer, players_dict, context):
        super().__init__(optimizer, players_dict, context)
        self.used_combinations = []  # type: List[Any]

    @property
    def max_repeating_players(self):
        optimizer = self.optimizer
        if optimizer.max_repeating_players:
            return optimizer.max_repeating_players
        return (self.player_pool.total_players or len(self.used_combinations[0])) - 1

    def apply(self, solver: Solver):
        for i, lineup in enumerate(self.context.exclude_lineups, start=1):
            variables = [self.players_dict[player] for player in lineup]
            solver.add_constraint(variables, None, SolverSign.LTE, self.max_repeating_players,
                                  name='exclude_lineups_%d' % i)

    def apply_for_iteration(self, solver, result):
        if not result:
            return
        self.used_combinations.append([self.players_dict[player] for player in result])
        for i, variables in enumerate(self.used_combinations, start=1):
            solver.add_constraint(variables, None, SolverSign.LTE, self.max_repeating_players,
                                  name='max_repeating_lineup_%d' % i)


class TotalPlayersRule(OptimizerRule):
    def apply(self, solver):
        total_players = self.player_pool.total_players
        if total_players:
            variables = self.players_dict.values()
            solver.add_constraint(variables, None, SolverSign.EQ, total_players, name='total_players')


class LineupBudgetRule(OptimizerRule):
    def apply(self, solver):
        if not self.optimizer.budget:
            return
        variables = []
        coefficients = []
        for player, variable in self.players_dict.items():
            variables.append(variable)
            coefficients.append(player.salary)
        solver.add_constraint(variables, coefficients, SolverSign.LTE, self.optimizer.budget, name='budget')


class LockedPlayersRule(OptimizerRule):
    def __init__(self, optimizer, players_dict, context):
        super().__init__(optimizer, players_dict, context)
        exposures = {}
        for player, variable in players_dict.items():
            exposures[variable.name] = player.max_exposure if player.max_exposure is not None \
                else self.context.max_exposure
        self.max_exposure_strategy = context.exposure_strategy(
            exposures, self.context.total_lineups)

    def apply_for_iteration(self, solver, result):
        force_variables = []
        exclude_variables = []
        removed_players = []
        for player, variable in self.players_dict.items():
            if not self.max_exposure_strategy.is_reached_exposure(variable.name):
                continue
            removed_players.append(player)
            exclude_variables.append(variable)
        for player in self.player_pool.locked_players:
            if player not in removed_players:
                force_variables.append(self.players_dict[player])
        if force_variables:
            solver.add_constraint(force_variables, None, SolverSign.EQ, len(force_variables), name='locked_players')
        if exclude_variables:
            solver.add_constraint(exclude_variables, None, SolverSign.EQ, 0, name='exclude_players')

    def post_optimize(self, solved_variables: List[str]):
        self.max_exposure_strategy.set_used(solved_variables)


class PositionsRule(OptimizerRule):
    def apply(self, solver):
        optimizer = self.optimizer
        if not self.player_pool.available_positions:
            return
        extra_positions = optimizer.players_with_same_position
        positions_combinations = set([tuple(sorted(player.positions)) for player in self.players_dict.keys()
                                      if len(player.positions) > 1])
        positions = get_positions_for_optimizer(self.player_pool.remaining_positions, positions_combinations)
        unique_positions = self.player_pool.available_positions
        players_by_positions = {
            position: {variable for player, variable in self.players_dict.items()
                       if position in player.positions and player not in self.player_pool.locked_players_with_positions
                       } for position in unique_positions
        }
        for position, places in positions.items():
            extra = 0
            if len(position) == 1:
                extra = extra_positions.get(position[0], 0)
            players_with_position = set()
            for pos in position:
                players_with_position.update(players_by_positions[pos])
            solver.add_constraint(players_with_position, None, SolverSign.GTE, places + extra,
                                  name='positions_%s' % '_'.join(position))


class TeamMatesRule(OptimizerRule):
    def apply(self, solver):
        for team, quantity in self.optimizer.players_from_one_team.items():
            players_from_same_team = [variable for player, variable in self.players_dict.items() if player.team == team]
            solver.add_constraint(players_from_same_team, None, SolverSign.EQ, quantity,
                                  name='players_from_one_team_%s' % team)


class MaxFromOneTeamRule(OptimizerRule):
    def apply(self, solver):
        if not self.optimizer.max_from_one_team:
            return
        for team in self.player_pool.available_teams:
            players_from_team = [variable for player, variable in self.players_dict.items() if player.team == team]
            solver.add_constraint(players_from_team, None, SolverSign.LTE, self.optimizer.max_from_one_team,
                                  name='max_from_one_team_%s' % team)


class MinSalaryCapRule(OptimizerRule):
    def apply(self, solver):
        variables = []
        coefficients = []
        for player, variable in self.players_dict.items():
            variables.append(variable)
            coefficients.append(player.salary)
        min_salary_cap = self.optimizer.min_salary_cap
        solver.add_constraint(variables, coefficients, SolverSign.GTE, min_salary_cap, name='min_salary_cap')


class ProjectedOwnershipRule(OptimizerRule):
    def apply(self, solver):
        min_variables = []
        min_coefficients = []
        max_variables = []
        max_coefficients = []
        min_projected_ownership = self.optimizer.min_projected_ownership
        max_projected_ownership = self.optimizer.max_projected_ownership
        for player, variable in self.players_dict.items():
            if player.projected_ownership is None:
                continue
            if min_projected_ownership:
                min_variables.append(variable)
                min_coefficients.append(player.projected_ownership - min_projected_ownership)
            if max_projected_ownership:
                max_variables.append(variable)
                max_coefficients.append(player.projected_ownership - max_projected_ownership)
        if min_variables:
            solver.add_constraint(min_variables, min_coefficients, SolverSign.GTE, 0, name='ownership_min')
        if max_variables:
            solver.add_constraint(max_variables, max_coefficients, SolverSign.LTE, 0, name='ownership_max')


class UniquePlayerRule(OptimizerRule):
    @staticmethod
    def sort_players(player_tuple):
        return player_tuple[0].full_name

    def apply(self, solver):
        data = sorted(self.players_dict.items(), key=self.sort_players)
        for player_id, group_iterator in groupby(data, key=self.sort_players):
            group = list(group_iterator)
            if len(group) == 1:
                continue
            variables = [variable for player, variable in group]
            solver.add_constraint(variables, None, SolverSign.LTE, 1, name='unique_player_%s' % player_id)


class LateSwapRule(OptimizerRule):
    def __init__(self, optimizer, players_dict, context):
        super().__init__(optimizer, players_dict, context)
        self.current_iteration = 0
        self.lineups = context.existed_lineups

    def apply_for_iteration(self, solver, result):
        current_lineup = self.lineups[self.current_iteration]
        unswappable_players = current_lineup.get_unswappable_players()
        remaining_positions = get_remaining_positions(self.optimizer.settings.positions, unswappable_players)
        # lock selected players
        for player in unswappable_players:
            solver.add_constraint([self.players_dict[player]], None, SolverSign.EQ, 1)
        # set remaining positions
        positions_combinations = set([tuple(sorted(player.positions)) for player in self.players_dict.keys()
                                      if len(player.positions) > 1])
        positions = get_positions_for_optimizer(remaining_positions, positions_combinations)
        players_for_optimization = set()
        for position, places in positions.items():
            players_with_position = [variable for player, variable in self.players_dict.items()
                                     if list_intersection(position, player.positions) and
                                     player not in unswappable_players]
            players_for_optimization.update(players_with_position)
            solver.add_constraint(players_with_position, None, SolverSign.GTE, places)
        # Set total players for optimization
        solver.add_constraint(players_for_optimization, None, SolverSign.EQ, len(remaining_positions))
        # Exclude players with active games
        for player, variable in self.players_dict.items():
            if player not in unswappable_players and player.is_game_started:
                solver.add_constraint([self.players_dict[player]], None, SolverSign.EQ, 0)
        self.current_iteration += 1


class GenericStacksRule(OptimizerRule):
    def __init__(self, optimizer, players_dict, context):
        super().__init__(optimizer, players_dict, context)
        self.stacks = list(chain.from_iterable(stack.build_stacks(context.players, optimizer)
                                               for stack in optimizer.stacks))
        self.with_exposures = any(stack.with_exposures for stack in self.stacks)
        exposures = {}
        for stack in self.stacks:
            for group in stack.groups:
                if group.max_exposure is None:
                    continue
                exposures[self._build_group_name(group)] = group.max_exposure
        self.exposure_strategy = context.exposure_strategy(
            exposures, self.context.total_lineups)

    def apply(self, solver):
        if not self.with_exposures:
            self._create_constraints(solver)

    def apply_for_iteration(self, solver, result):
        if self.with_exposures:
            self._create_constraints(solver)

    @staticmethod
    def _build_group_name(group: 'BaseGroup'):
        return 'stack_%s' % group.uuid.hex

    def _create_constraints(self, solver: Solver,):
        players_in_stack = defaultdict(set)  # type: Dict[Player, Set[Any]]
        for stack in self.stacks:
            combinations_variables = {}
            for group in stack.groups:
                group_name = self._build_group_name(group)
                sub_groups = group.get_all_players_groups()
                if self.exposure_strategy.is_reached_exposure(group_name) or \
                        (group.parent and self.exposure_strategy.is_reached_exposure(
                            self._build_group_name(group.parent))):
                    for sub_group in sub_groups:
                        if sub_group.max_from_group is None:
                            continue
                        solver.add_constraint([self.players_dict[p] for p in sub_group.players], None, SolverSign.EQ, 0)
                    max_group = sorted(sub_groups, key=lambda g: g.min_from_group)[0]  # type: ignore
                    if max_group.min_from_group:
                        solver.add_constraint([self.players_dict[p] for p in max_group.players], None, SolverSign.LTE,
                                              max_group.min_from_group - 1)
                    continue
                solver_variable = None
                if any(sub_group.min_from_group is not None for sub_group in sub_groups):
                    solver_variable = solver.add_variable(group_name)
                    if group.depends_on is None:
                        combinations_variables[group_name] = solver_variable
                for sub_group in sub_groups:
                    variables = [self.players_dict[p] for p in sub_group.players if p in self.players_dict]
                    if sub_group.min_from_group is not None:
                        if not stack.can_intersect:
                            for player in sub_group.players:
                                players_in_stack[player].add(solver_variable)
                        solver.add_constraint(variables, None, SolverSign.GTE, sub_group.min_from_group * solver_variable)
                        if sub_group.max_from_group is not None:
                            solver.add_constraint(variables, None, SolverSign.LTE, sub_group.max_from_group)
                    elif sub_group.max_from_group is not None:
                        solver_variable = solver.add_variable(group_name, min_value=0, max_value=sub_group.max_from_group)
                        solver.add_constraint(variables, None, SolverSign.EQ, solver_variable)
                    if group.depends_on is not None:
                        total_players_var = solver.add_variable('total_players_%s' % group_name, min_value=0,
                                                                max_value=sub_group.max_from_group or len(variables))
                        solver.add_constraint(variables, None, SolverSign.EQ, total_players_var)
                        depend_var = self.players_dict.get(group.depends_on, 0)
                        solver.add_constraint(
                            [total_players_var], None, SolverSign.GTE,
                            depend_var * (sub_group.min_from_group or 1)
                        )
                        if group.strict_depend:
                            solver.add_constraint(
                                [total_players_var], None, SolverSign.LTE,
                                depend_var * (sub_group.max_from_group or len(variables))
                            )
            if combinations_variables:
                solver.add_constraint(combinations_variables.values(), None, SolverSign.GTE, 1)
        for player, stacks_vars in players_in_stack.items():
            if len(stacks_vars) > 1:
                solver.add_constraint(stacks_vars, None, SolverSign.LTE, 1)

    def post_optimize(self, solved_variables):
        self.exposure_strategy.set_used(solved_variables)


class MinExposureRule(OptimizerRule):
    def __init__(self, optimizer, players_dict, context):
        super().__init__(optimizer, players_dict, context)
        self.min_exposure_players = {
            player: round(player.min_exposure * context.total_lineups)
            for player in context.players if player.min_exposure
        }
        self.positions = {}  # type: Dict[Tuple[str, ...], int]
        if self.min_exposure_players:
            self.positions = get_positions_for_optimizer(optimizer.settings.positions, None)

    def apply_for_iteration(self, solver, result):
        if not self.min_exposure_players:
            return
        if result:
            for player in result:
                if player not in self.min_exposure_players:
                    continue
                self.min_exposure_players[player] -= 1
                if self.min_exposure_players[player] == 0:
                    del self.min_exposure_players[player]
        self._create_constraints(solver)

    def _create_constraints(self, solver: Solver) -> None:
        remaining_lineups = self.context.remaining_lineups
        for positions, total_for_positions in self.positions.items():
            min_exposure_players = [p for p in self.min_exposure_players if list_intersection(p.positions, positions)]
            if not min_exposure_players:
                continue
            total_force = sum(self.min_exposure_players[p] for p in min_exposure_players) - \
                          total_for_positions * (remaining_lineups - 1)
            if total_force > 0:
                variables = [self.players_dict[p] for p in min_exposure_players]
                total_force = min(total_force, ceil(total_force / remaining_lineups))
                solver.add_constraint(variables, None, SolverSign.GTE, total_force)
        for player, total_lineups in self.min_exposure_players.items():
            if total_lineups >= remaining_lineups:
                solver.add_constraint([self.players_dict[player]], None, SolverSign.EQ, 1)


class RestrictPositionsForOpposingTeam(OptimizerRule):
    MULTIPLIER = 1000

    def apply(self, solver):
        if not self.optimizer.opposing_teams_position_restriction:
            return
        for game in self.optimizer.player_pool.games:
            home_team_players = {player: variable for player, variable in self.players_dict.items()
                                  if player.team == game.home_team}
            away_team_players = {player: variable for player, variable in self.players_dict.items()
                                   if player.team == game.away_team}
            first_team_positions, second_team_positions = self.optimizer.opposing_teams_position_restriction
            for first_team_players, second_team_players in permutations([home_team_players, away_team_players], 2):
                first_team_variables = [variable for player, variable in first_team_players.items()
                                        if list_intersection(player.positions, first_team_positions)]
                second_team_variables = [variable for player, variable in second_team_players.items()
                                         if list_intersection(player.positions, second_team_positions)]
                aggregated_var = solver.add_variable(str(uuid4()), min_value=0, max_value=len(second_team_variables))
                solver.add_constraint(second_team_variables, None, SolverSign.EQ, aggregated_var)
                for var in first_team_variables:
                    solver.add_constraint([var, aggregated_var], [self.MULTIPLIER, 1],
                                          SolverSign.LTE, self.MULTIPLIER + self.optimizer.opposing_teams_max_allowed)


class RestrictPositionsForSameTeamRule(OptimizerRule):
    def apply(self, solver):
        all_restrict_positions = self.optimizer.same_team_restrict_positions
        if not all_restrict_positions:
            return
        players_by_team = get_players_grouped_by_teams(self.players_dict.keys())
        for restrict_positions in all_restrict_positions:
            for team_players in players_by_team.values():
                first_position_players = [player for player in team_players
                                          if restrict_positions[0] in player.positions]
                second_position_players = [player for player in team_players
                                           if restrict_positions[1] in player.positions]
                for players_combination in product(first_position_players, second_position_players):
                    if players_combination[0] == players_combination[1]:
                        continue
                    variables = [self.players_dict[player] for player in players_combination]
                    solver.add_constraint(variables, None, SolverSign.LTE, 1)


class ForcePositionsForOpposingTeamRule(OptimizerRule):
    def apply(self, solver):
        raw_all_force_positions = self.optimizer.opposing_team_force_positions
        if not raw_all_force_positions:
            return
        all_force_positions = [tuple(sorted(positions)) for positions in raw_all_force_positions]
        for positions, total_combinations in Counter(all_force_positions).items():
            positions_vars = []
            combinations_count = 0
            for game in self.player_pool.games:
                first_team_players = {player: variable for player, variable in self.players_dict.items()
                                      if player.team == game.home_team}
                second_team_players = {player: variable for player, variable in self.players_dict.items()
                                       if player.team == game.away_team}
                for first_team_positions, second_team_positions in permutations(positions, 2):
                    first_team_variables = [variable for player, variable in first_team_players.items()
                                            if first_team_positions in player.positions]
                    second_team_variables = [variable for player, variable in second_team_players.items()
                                             if second_team_positions in player.positions]
                    for variables in product(first_team_variables, second_team_variables):
                        solver_variable = solver.add_variable('force_positions_%s_%d' % (positions, combinations_count))
                        combinations_count += 1
                        positions_vars.append(solver_variable)
                        solver.add_constraint(variables, None, SolverSign.GTE, 2 * solver_variable)
            solver.add_constraint(positions_vars, None, SolverSign.GTE, total_combinations)


class RosterSpacingRule(OptimizerRule):
    def apply(self, solver):
        optimizer = self.optimizer
        positions, spacing = optimizer.spacing_positions, optimizer.spacing
        if not spacing or not positions:
            return
        players_by_roster_positions = defaultdict(list)  # type: Dict[int, List[Tuple[Player, Any]]]
        for player, variable in self.players_dict.items():
            if player.roster_order is None or not list_intersection(player.positions, positions):
                continue
            players_by_roster_positions[player.roster_order].append((player, variable))
        max_order = max(players_by_roster_positions.keys())
        for roster_position, players in players_by_roster_positions.items():
            restricted_players = chain.from_iterable(
                players for players_spacing, players in players_by_roster_positions.items()
                if max_order - spacing + roster_position >= players_spacing >= roster_position + spacing
            )
            for first_player, first_variable in restricted_players:
                for second_player, second_variable in players:
                    if first_player.team != second_player.team:
                        continue
                    solver.add_constraint([first_variable, second_variable], None, SolverSign.LTE, 1)


class FanduelBaseballRosterRule(OptimizerRule):
    HITTERS = ('1B', '2B', '3B', 'SS', 'C', 'OF')
    MAXIMUM_HITTERS_FROM_ONE_TEAM = 4

    def apply(self, solver):
        players_dict = {player: variable for player, variable in self.players_dict.items() if
                        list_intersection(player.positions, self.HITTERS)}
        for team in self.optimizer.player_pool.available_teams:
            players_from_team = [variable for player, variable in players_dict.items() if player.team == team]
            solver.add_constraint(players_from_team, None, SolverSign.LTE, self.MAXIMUM_HITTERS_FROM_ONE_TEAM,
                                  name='fanduel_max_hitters_%s' % team)


class TotalTeamsRule(OptimizerRule):
    def apply(self, solver):
        settings = self.optimizer.settings
        min_teams = self.optimizer.min_teams or settings.min_teams
        max_teams = self.optimizer.max_teams
        if not min_teams and not max_teams:
            return
        total_players = self.optimizer.settings.get_total_players()
        players_by_teams = get_players_grouped_by_teams(self.players_dict.keys(), for_positions=[
            position for position in self.optimizer.player_pool.available_positions
            if position not in settings.total_teams_exclude_positions
        ])
        teams_variables = []
        for team, team_players in players_by_teams.items():
            variable = solver.add_variable('total_teams_%s' % team)
            teams_variables.append(variable)
            variables = [self.players_dict[player] for player in team_players]
            solver.add_constraint(variables, None, SolverSign.LTE, variable * total_players)
            solver.add_constraint(variables, None, SolverSign.GTE, variable)
        if min_teams == max_teams:
            solver.add_constraint(teams_variables, None, SolverSign.EQ, min_teams, name='exact_teams')
        if min_teams:
            solver.add_constraint(teams_variables, None, SolverSign.GTE, min_teams, name='min_teams')
        if max_teams:
            solver.add_constraint(teams_variables, None, SolverSign.LTE, max_teams, name='max_teams')


class FanduelSingleGameMaxQBRule(OptimizerRule):
    def apply(self, solver):
        variables = [var for player, var in self.players_dict.items() if 'QB' in player.original_positions]
        solver.add_constraint(variables, None, SolverSign.LTE, 2, name='fanduel_single_game_max_qb')


class MinStartersRule(OptimizerRule):
    def apply(self, solver):
        min_starters = self.optimizer.min_starters
        if not min_starters:
            return
        variables = [variable for player, variable in self.players_dict.items() if player.is_confirmed_starter]
        solver.add_constraint(variables, None, SolverSign.GTE, min_starters, name='min_starters')


class MinGamesRule(OptimizerRule):
    def apply(self, solver):
        min_games = self.optimizer.settings.min_games
        if not min_games:
            return
        total_players = self.optimizer.settings.get_total_players() or 100
        players_by_games = defaultdict(list)  # type: DefaultDict[GameInfo, List[Player]]
        for player in self.player_pool.filtered_players:
            if player.game_info:
                players_by_games[player.game_info].append(player)
        game_variables = []
        for game, game_players in players_by_games.items():
            variable = solver.add_variable('total_game_%s_%s' % (game.home_team, game.away_team))
            game_variables.append(variable)
            variables = [self.players_dict[player] for player in game_players]
            solver.add_constraint(variables, None, SolverSign.LTE, variable * total_players)
            solver.add_constraint(variables, None, SolverSign.GTE, variable)
        if len(game_variables) >= min_games:
            solver.add_constraint(game_variables, None, SolverSign.GTE, min_games, name='min_games')


class DraftKingsBaseballRosterRule(OptimizerRule):
    HITTERS = ('1B', '2B', '3B', 'SS', 'C', 'OF')
    MAXIMUM_HITTERS_FROM_ONE_TEAM = 5

    def apply(self, solver):
        players_dict = {player: variable for player, variable in self.players_dict.items() if
                        list_intersection(player.positions, self.HITTERS)}
        for team in self.player_pool.available_teams:
            players_from_team = [variable for player, variable in players_dict.items() if player.team == team]
            solver.add_constraint(players_from_team, None, SolverSign.LTE, self.MAXIMUM_HITTERS_FROM_ONE_TEAM,
                                  name='dk_max_hitters_%s' % team)


class DraftKingsTiersRule(OptimizerRule):
    @staticmethod
    def sort_player(player):
        return player.positions[0]

    def apply(self, solver):
        for tier, players in groupby(sorted(self.players_dict.keys(), key=self.sort_player), key=self.sort_player):
            variables = [self.players_dict[player] for player in players]
            solver.add_constraint(variables, None, SolverSign.EQ, 1, name='dk_tier_%s' % tier)


class TeamsExposureRule(OptimizerRule):
    def __init__(self, optimizer: 'LineupOptimizer', players_dict: Dict[Player, Any], context: OptimizationContext):
        super().__init__(optimizer, players_dict, context)
        exposures = {}
        for team in self.player_pool.available_teams:
            team_exposure = (optimizer.teams_exposures or {}).get(team, optimizer.default_team_exposure)
            if team_exposure is not None:
                exposures[team] = team_exposure
        exposure_strategy = optimizer.teams_exposure_strategy or context.exposure_strategy
        self.max_exposure_strategy = exposure_strategy(exposures, self.context.total_lineups)
        variables_by_team = defaultdict(list)
        for player, variable in players_dict.items():
            variables_by_team[player.team].append(variable)
        self.variables_by_team = variables_by_team

    def apply_for_iteration(self, solver, result):
        if result is None:
            return
        used_teams = list({p.team for p in result})
        self.max_exposure_strategy.set_used(used_teams)
        for team in self.player_pool.available_teams:
            if not self.max_exposure_strategy.is_reached_exposure(team):
                continue
            team_variables = self.variables_by_team.get(team)
            if not team_variables:
                continue
            solver.add_constraint(team_variables, None, SolverSign.EQ, 0)



from collections import defaultdict
from typing import Dict, List


class BaseExposureStrategy:
    def __init__(self, exposures: Dict[str, float], total_lineups: int) -> None:
        self.exposures = exposures
        self.total_lineups = total_lineups
        self.used_vars = defaultdict(int)  # type: Dict[str, int]

    def set_used(self, variables: List[str]):
        for var in variables:
            if var in self.exposures:
                self.used_vars[var] += 1

    def is_reached_exposure(self, var: str) -> bool:
        raise NotImplementedError


class TotalExposureStrategy(BaseExposureStrategy):
    def is_reached_exposure(self, var):
        max_exposure = self.exposures.get(var)
        if not max_exposure:
            return False
        used = self.used_vars.get(var, 0)
        return max_exposure <= used / self.total_lineups


class AfterEachExposureStrategy(BaseExposureStrategy):
    def __init__(self, exposures, total_lineups):
        super().__init__(exposures, total_lineups)
        self.current_iteration = 0

    def set_used(self, variables: List[str]):
        super().set_used(variables)
        self.current_iteration += 1

    def is_reached_exposure(self, var):
        max_exposure = self.exposures.get(var)
        if not max_exposure or not self.current_iteration:
            return False
        used = self.used_vars.get(var, 0)
        return max_exposure <= used / self.current_iteration
